/* Copyright 2009-2016 David Hadka
 *
 * This file is part of the MOEA Framework.
 *
 * The MOEA Framework is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The MOEA Framework is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the MOEA Framework.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.moeaframework.algorithm;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import org.moeaframework.core.FastNondominatedSorting;
import org.moeaframework.core.FitnessEvaluator;
import org.moeaframework.core.Initialization;
import org.moeaframework.core.NondominatedPopulation;
import org.moeaframework.core.NondominatedSorting;
import org.moeaframework.core.Population;
import org.moeaframework.core.Problem;
import org.moeaframework.core.Selection;
import org.moeaframework.core.Solution;
import org.moeaframework.core.Variation;
import org.moeaframework.core.comparator.ChainedComparator;
import org.moeaframework.core.comparator.FitnessComparator;
import org.moeaframework.core.comparator.NondominatedSortingComparator;
import org.moeaframework.core.comparator.ObjectiveComparator;
import org.moeaframework.core.comparator.RankComparator;
import org.moeaframework.core.indicator.Hypervolume;
import org.moeaframework.core.operator.TournamentSelection;

/**
 * Implementation of the S-metric Selection MOEA (SMS-MOEA).  The S metric is
 * also known as the hypervolume indicator.
 * <p>
 * References:
 * <ol>
 *   <li>Emmerich, M., N. Beume, and B. Naujoks (2007).  An EMO Algorithm Using
 *       the Hypervolume Measure as Selection Criterion.  European Journal of
 *       Operational Research, 3:1653-1669.
 */
public class DFSMSEMOA extends AbstractEvolutionaryAlgorithm {
	
	/**
	 * The fitness evaluator to use (e.g., hypervolume or additive-epsilon
	 * indicator).
	 */
	private FitnessEvaluator fitnessEvaluator;
	
	/**
	 * The selection operator.
	 */
	private Selection selection;
	
	/**
	 * The variation operator.
	 */
	private Variation variation;
	
	public static final String RANK_ATTRIBUTE = "rank";
	public static final String CROWDING_NUMBER = "INDEX";
	public static final String FITNESS_ATTRIBUTE = "fitness";
	// control the shape of Harrington's two-sided DF, the higher, the more it looks like step function
	private int n=4;
	private double epsilon=1.1e-16;
   private Problem pp;
	/**
	 * Constructs a new DF-SMS-EMOA instance.
	 * 
	 * @param problem the problem
	 * @param initialization the initialization operator
	 * @param variation the variation operator
	 * @param fitnessEvaluator the fitness evaluator
	 */
	public DFSMSEMOA(Problem problem, Initialization initialization,
			Variation variation, FitnessEvaluator fitnessEvaluator) {
		super(problem,
				new Population(),
				null,
				initialization);
		this.variation = variation;
		this.fitnessEvaluator = fitnessEvaluator;
		this.pp=problem;
		if (fitnessEvaluator ==  null) {
			selection = new TournamentSelection(
					new NondominatedSortingComparator());
		} else {
			selection = new TournamentSelection(
					new NondominatedFitnessComparator());
		}
		
		
	}

	@Override
	protected void initialize() {
		super.initialize();
				
// DF coordinate transformation	
		Population p=DFTransformation(population);
		
		new NondominatedSorting().evaluate(p);
		
		fitnessEvaluator.evaluateDF(p);
		
		// get rank and fitness from p
		for(int i=0;i<population.size();i++){
			if((int)population.get(i).getAttribute(CROWDING_NUMBER)==(int)p.get(i).getAttribute(CROWDING_NUMBER))
			{	
			int rank=(int)p.get(i).getAttribute(RANK_ATTRIBUTE);
			population.get(i).setAttribute(RANK_ATTRIBUTE, rank);
			
			double fit=(double)p.get(i).getAttribute(FITNESS_ATTRIBUTE);
			population.get(i).setAttribute(FITNESS_ATTRIBUTE, fit);
			}
			else
			{
				System.out.println("population order changed!");
			}
		}
			
		//for test
	//	int num=population.size();
		
	}

	private Population DFTransformation(Population population) {
		String DFshape="OneSide";// "OneSide"
		Population newPop=new Population();		
		double[] lsl={0.0,0.6};
		double[] uslnew={0.3,1.0};
//		double[] lsl={0.0,1};
//		double[] uslnew={0.3,0.6};
		// test find number of solutions in ROI
//		int num=0;
//		for(Solution s: population){
//			boolean isIn=true;
//			for(int j=0;j<population.get(0).getNumberOfObjectives();j++){
//				if(s.getObjective(j)<lsl[j]||s.getObjective(j)>uslnew[j]){
//					isIn=false;
//					break;
//				}
//			}
//			if(isIn){
//				num++;
//			}			
//		}
//		if(num==population.size())
//		{
//			num++; 
//		}
//		//test find number of solutions in ROI end
		
		int index=0;
	
		int numberOfObjectives = population.get(0).getNumberOfObjectives();
		double[] d=new double[numberOfObjectives];
		for(int i=0;i<numberOfObjectives;i++){
			d[i]=0.001;
		}
		
		// epsiloncheck
		for(int k=0;k<numberOfObjectives;k++){
			population.sort(new ObjectiveComparator(k));
			double median=0;
			int PS=population.size();
			if (PS % 2 == 0 ){
				median = ((double)population.get(PS/2).getObjective(k) + (double)population.get(PS/2 - 1).getObjective(k))/2;
			    }else if(PS>1){
			    	 median = (double)population.get(PS/2).getObjective(k);
			    }else if(PS==1)
			    {
			    	median=(double)population.get(0).getObjective(k);
			    }
			if(DFshape=="OneSide"){
			double dmedian= OneSideDFFunction(median, lsl[k], uslnew[k],d[k]);  
//			double dmedian= TwoSideDFFunction(median, lsl[k], uslnew[k]);
			if(dmedian<epsilon){
				uslnew[k]=median;
				d[k]=epsilon;			
			}		
			}// oneside
			else if(DFshape=="TwoSide"){
				
				double dmedian= TwoSideDFFunction(median, lsl[k], uslnew[k]);
				if(dmedian<epsilon){
					if(median>uslnew[k]){
						uslnew[k]=median;
					}
					else if(median<lsl[k]){
						lsl[k]=median;
					}
				}
			}// two side
		}// number of objectives
		if(DFshape=="TwoSide"){
		
				for(Solution s:population){
					s.setAttribute(CROWDING_NUMBER, index);
					index++;
					Solution solutiontemp=TwoSideDFTransformation( s,  lsl, uslnew);
					newPop.add(solutiontemp);
				}
			
		}
		else if(DFshape=="OneSide"){
		for(Solution s:population){
			s.setAttribute(CROWDING_NUMBER, index);
			index++;
			Solution solutiontemp=OneSideDFTransformation( s,  lsl, uslnew, d);
			newPop.add(solutiontemp);
		}
		}
		
		
		
		
		return newPop;
	}



	@Override
	protected void iterate() {
		int populationSize = population.size();
		Solution[] parents = selection.select(variation.getArity(), population);
		Solution[] offspring = variation.evolve(parents);
		
		evaluate(offspring[0]);
		population.add(offspring[0]);
		
	    Population p=DFTransformation(population);
		//rank
		new NondominatedSorting().evaluate(p);
		
       
		computeFitnessForLastFront(p);
		
		// get rank and fitness from p
		for(int i=0;i<population.size();i++){
			if((int)population.get(i).getAttribute(CROWDING_NUMBER)==(int)p.get(i).getAttribute(CROWDING_NUMBER))
			{	int rank=(int)p.get(i).getAttribute(RANK_ATTRIBUTE);
			population.get(i).setAttribute(RANK_ATTRIBUTE, rank);
			
			double fit=(double)p.get(i).getAttribute(FITNESS_ATTRIBUTE);
			population.get(i).setAttribute(FITNESS_ATTRIBUTE, fit);
			}
			else
			{
				System.out.println("population order changed!");
			}
		}
		
			
		population.truncate(populationSize, 
					new NondominatedFitnessComparator());
		
	}
	
	/**
	 * Computes the fitness for the last front.
	 */
	private void computeFitnessForLastFront(Population p) {
		Population front = new Population();
		int rank = 0;
		
		for (Solution solution : p) {
			int solutionRank = (Integer)solution.getAttribute(
					FastNondominatedSorting.RANK_ATTRIBUTE);
			
			if (solutionRank > rank) {
				front.clear();
				rank = solutionRank;
			}
			
			if (solutionRank == rank) {
				front.add(solution);
			}
			
			solution.setAttribute(FitnessEvaluator.FITNESS_ATTRIBUTE, 0.0);
		}
		
		fitnessEvaluator.evaluateDF(front);
	}
	
	private class NondominatedFitnessComparator extends ChainedComparator
	implements Comparator<Solution>, Serializable {

		private static final long serialVersionUID = -4088873047790962685L;

		public NondominatedFitnessComparator() {
			super(new RankComparator(), new FitnessComparator(
					fitnessEvaluator.areLargerValuesPreferred()));
		}

	}
	

	private Solution OneSideDFTransformation(Solution s, double[] lsl, double[] usl, double[] d) {
		Solution newSolution = s.copy();
		
		
		int tempindex = (int)s.getAttribute(CROWDING_NUMBER);
		newSolution.setAttribute(CROWDING_NUMBER, tempindex);
//
//		int rank= (int) s.getAttribute(RANK_ATTRIBUTE);
//		newSolution.setAttribute(RANK_ATTRIBUTE, rank);
		
		for (int i = 0; i < s.getNumberOfObjectives(); i++)
		{
//			newSolution.setObjective(i, -TwoSideDFFunction(s.getObjective(i),  lsl[i], usl[i]));
			newSolution.setObjective(i, -OneSideDFFunction(s.getObjective(i), lsl[i], usl[i],d[i]));
		}
				
	return newSolution;
	}
	
	private Solution TwoSideDFTransformation(Solution s, double[] lsl, double[] uslnew) {
        Solution newSolution = s.copy();	
		int tempindex = (int)s.getAttribute(CROWDING_NUMBER);
		newSolution.setAttribute(CROWDING_NUMBER, tempindex);
		for (int i = 0; i < s.getNumberOfObjectives(); i++)
		{
			newSolution.setObjective(i, -TwoSideDFFunction(s.getObjective(i), lsl[i], uslnew[i]));	
		}
				
	return newSolution;
	
		
	}
	
	//Harrington's one-sided DF	
private double OneSideDFFunction(double objective,  double lower, double upper, double d) {
		double result=0;
		double y1=lower;
		double d1=0.999;
		double y2=upper;
		double d2=d;
		double b1=(-Math.log(-Math.log(d2))+ Math.log(-Math.log(d1)))/(y2-y1);
		double b0=-Math.log(-Math.log(d1))-b1*y1;
		result=Math.exp(-Math.exp(-(b0+b1*objective)));
		
		return result;
	}

	//Harrington's two-sided DF
	private double TwoSideDFFunction(double objective, double lsl, double usl) {
		double result=0;
		double y=(2*objective-(lsl+usl))/(usl-lsl);
		result=Math.exp(-Math.pow(Math.abs(y),n));
		
		return result;
	}

	@Override
	public Population getCurrentPopulation() {
		// TODO Auto-generated method stub
		return null;
	}
	

}
