/* Copyright 2009-2016 David Hadka
 *
 * This file is part of the MOEA Framework.
 *
 * The MOEA Framework is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The MOEA Framework is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the MOEA Framework.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.moeaframework.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.moeaframework.Executor;
import org.moeaframework.core.NondominatedPopulation.DuplicateMode;
//import org.moeaframework.core.comparator.CrowdingComparator1;
//import org.moeaframework.core.comparator.CrowdingComparator2;
import org.moeaframework.core.comparator.DominanceComparator;
import org.moeaframework.core.comparator.FitnessComparator1;
import org.moeaframework.core.comparator.FitnessComparator2;
import org.moeaframework.core.comparator.ObjectiveComparator;
import org.moeaframework.core.comparator.ParetoDominanceComparator;
import org.moeaframework.core.comparator.TchebycheffDistanceComparator1;
import org.moeaframework.core.comparator.TchebycheffDistanceComparator2;

/**
 * Non-dominated sorting algorithm for dominance depth ranking. Assigns the
 * {@code rank} and {@code crowdingDistance} attributes to solutions. Solutions
 * of rank 0 belong to the Pareto non-dominated front.
 * <p>
 * Despite its name, this naive non-dominated sort implementation tends to be
 * faster than the "fast non-dominated sort" implementation from [1].  This
 * is primarily due to the fact that for the average case, the "fast" version
 * always requires O(MN^2) comparisons while this naive implementations requires
 * only (K-1)/2 * M * (N-1)*N/2, assuming there are K equally sized fronts.
 * <p>
 * References:
 * <ol>
 * <li>Deb et al (2002). "A Fast and Elitist Multiobjective Genetic Algorithm:
 * NSGA-II." IEEE Transactions on Evolutionary Computation. 6(2):182-197.
 * </ol>
 */
public class NondominatedSorting {

	/**
	 * Attribute key for the rank of a solution.
	 */
	public static final String RANK_ATTRIBUTE = "rank";

	/**
	 * Attribute key for the crowding distance of a solution.
	 */
	public static final String CROWDING_ATTRIBUTE = "crowdingDistance";
	public static final String CROWDING_Distance = "RealCrowdingDistance";
	// convert to double arrays
//	public static final String CROWDING1_ATTRIBUTE = "crowdingDistance1";
//	public static final String CROWDING2_ATTRIBUTE = "crowdingDistance2";
	
	public static final String CROWDING_NUMBER = "INDEX";
	
	public static final String DISTANCE_ATTRIBUTE = "TchebycheffDistance";
	
	public static final String SORT_ATTRIBUTE = "SortingCriterion"; // the attribute to sort solutions, can be crowdingdistance1, crowdingDistance2, Distance1_ATTRIBUTE,Distance2_ATTRIBUTE
	// convert to double arrays
//	public static final String Distance1_ATTRIBUTE = "TchebycheffDistance1";
//	public static final String Distance2_ATTRIBUTE = "TchebycheffDistance2";
//	public static final String LABEL1_ATTRIBUTE = "LABEL1";
//	public static final String LABEL2_ATTRIBUTE = "LABEL2";
	
	// added by longmei li
	
	/**
	 * Attribute key for the Euclidean distance to RP of a solution.
	 */
	public static final String EDistance_ATTRIBUTE = "WeightedEuclideanDistanceToRP";
	
	/**
	 * Attribute key for the Chebycheff ASF of a solution.
	 */
	public static final String ChebyASF_ATTRIBUTE = "ChebycheffASFValue";
	
	/**
	 * Attribute key for the Chebycheff distance of a solution.
	 */
	public static final String ChebyDis_ATTRIBUTE = "ChebycheffDistance";
	
	// r-NSGA-II
	public static  double MaxDistance=0;// Max WeightedEuclideanDistanceToRP in this population
	public static  double MinDistance=Double.POSITIVE_INFINITY;// Min WeightedEuclideanDistanceToRP in this population	
	//CP-NSGA-II
	public static  double MaxChebyASF=0;// Max ChebycheffASFValue in this population
	public static  double MinChebyASF=Double.POSITIVE_INFINITY;// Min ChebycheffASFValue in this population	

	
	// epsilon clearing parameter
//	public static double epsilon=0.0002;
	public static double epsilon=0.00001;
	//used by both Distance_Attribute, Chebycheve_Attribute and weight function
//	private double[] referencepoint={-1,-0.95};
	public static double[][] bounds=new double[10][2];
	private double[] weights={0.5,0.5}; // should be same dimension as NumOfObj
//	private double[] weights={0.333,0.333,0.333};	
	
	// use a class to describe the preferences, used for TMOEAs and R-NSGA-II
	public static final class preference{
		public static String  shape="rectangle"; // shape can be "rectangle","circle", "point"
//		public static Double[][] regions={{0.1, 0.8, 0.1},{0.8, 0.1, 0.1}};//  format for circle: {CenterPoint_X, CenterPoint_Y,... Radius},{},...	
//		public static Double[][] regions={{0.0,0.6, 0.3,1.0},{0.6,0.0,1.0,0.3}};//format for rectangle: {LowerBound_X, LowerBound_Y,...UpperBound_X, UpperBound_Y},{}...
//		public static Double[][] regions={{0.1, 0.4},{0.7,0.4},{0.4, 0.367544},{0.9, 0.1}};//format for point: {{x,y,z...},{},...}		
		public static int NumOfROI=2; // number of preference regions
// 2D-benchmark problem
//		public static Double[][] regions={{0.0,0.7, 0.2,1.0},{0.8,0.0,1.0,0.2}}; //ZDT1
//		public static Double[][] regions={{0.3,0.3, 0.5,0.5}}; //ZDT1
//		public static Double[][] regions={{0.6,0.5,0.8,0.7}}; //ZDT2
//		public static Double[][] regions={{0.0,0.9,0.2,1.0},{0.8,0.0,1.0,0.2}}; //ZDT2
//		public static Double[][] regions={{0.0,0.6,0.3,0.8},{0.6,-0.6,0.9,-0.3}};//ZDT3
//		public static Double[][] regions={{0.2,0.0, 0.7,0.4}}; //ZDT3
//		public static Double[][] regions={{0.3,0.3,0.5,0.5}}; //ZDT4
//		public static Double[][] regions={{0.0,0.7,0.3,1.0},{0.7,0.0,1.0,0.3}}; //ZDT4

// 3D-benchmark problem
//		public static Double[][] regions={{0.0,0.1,0.1,0.2,0.3,0.3}};//DTLZ1_3
//		public static Double[][] regions={{0.13,0.13,0.07,0.28,0.28,0.1},{0.04,0.16,0.23,0.12,0.24,0.25},{0.17,0.03,0.23,0.25,0.11,0.25}};//DTLZ1_3 smile
//		public static Double[][] regions={{0.0,0.0,1.0,0.1},{0.0,1.0,0.0,0.1},{1.0,0.0,0.0,0.1}};//DTLZ2_3, 3 balls
//		public static Double[][] regions={{0.6,0.6,0.6,0.2}};//DTLZ2_3, one ball
//		public static Double[][] regions={{0.5,0.5,0.5, 0.7,0.7,0.7}};//DTLZ3_3		
//		public static Double[][] regions={{0.3,0.3,0.8,0.4,0.4,1.0},{0.6,0.6,0.2,0.8,0.8,0.3}};//DTLZ3_3	
// satellite scheduling problem		
//		public static Double[][] regions={{-0.6,  -0.8}};//Sat
//		public static Double[][] regions={{-1.0, -1.0, -0.95,-0.9}};//Sat2D
		public static Double[][] regions={{-0.6, -1.0, -0.8,-0.5,-0.9,-0.4},{-0.5,-1.0,-1.0,-0.2,-0.90,-0.90}};//Sat3D
		
//		public static Double[][] regions={{0.7, 0.8}};//ZDT2-point
//		public static Double[][] regions={{0.8,0.1,0.2},{0.23834163827671956, 0.15557436938154254, 0.9586395981050985}};//ZDT1-point
//		public static Double[][] regions={{0.1, 0.6},{0.55, 0.0},{0.8,-0.5}};// ZDT3
//		public static Double[][] regions={{0.2, 0.2, 0.8},{0.8,0.8,0.3}};

//		public static Double[][] regions={{0.8, 0.0, 0.95, 0.65},{0.4, 0.0, 0.5,0.65},{0.15,0.0,0.2,0.65}};// ZDT1-vs-pNSGAII
//		public static Double[][] regions={{0.1, 0.6},{0.3, 0.6},{0.5,0.2},{0.7,0.2},{0.9,0.0}};//format for point: X,Y,...

//		public static Double[][] regions={{0.1, 0.4},{0.7,0.4},{0.4, 0.367544},{0.9, 0.1}};

//    	public static Double[][] regions={{0.3,0.3, 0.6,0.6},{0.4,0.2,0.8,0.6}}; // two overlapping rectangles
//		public static Double[][] regions={{0.0,0.6, 0.2,0.8},{0.3,0.3,0.6,0.5},{0.7,0.0,1.0,0.2}};

//	
//		public static Double[][] regions={{0.0,0.8, 0.15,1.0},{0.5,0.65,0.6,0.8},{0.85,0.0,1.0,0.2}};//ZDT2
//		public static Double[][] regions={{0.5,0.5,0.8,0.8}};// inadequate /feasible solution 
//		public static Double[][] regions={{0.08,0.2, 0.18,0.4},{0.7,-0.4,0.8,-0.2}};//ZDT3
//		public static Double[][] regions={{0.2,0.2,0.6},{0.8,0.6,1.0}};//DTLZ2
//		public static Double[][] regions={{0.17,0.08,0.1,0.29,0.2,0.22},{0.12,0.12,0.12,0.24,0.24,0.24}};
//		public static Double[][] regions={{0.1,0.8,0.2,0.9},{0.8,0.1,0.9,0.2}}; // DTLZ2-vsASEMOA
//		public static Double[][] regions={{0.15,0.15,0.21,0.21}};// DTLZ2-vsASEMOA
//		
		
		
	};
	
	///////////////////////////////////////////////longmei//////////////////////////////////////////////
	
	/**
	 * The dominance comparator.
	 */
	protected final DominanceComparator comparator;

	/**
	 * Constructs a fast non-dominated sorting operator using Pareto dominance.
	 */
	public NondominatedSorting() {
		this(new ParetoDominanceComparator());
	}

	/**
	 * Constructs a non-dominated sorting operator using the specified
	 * dominance comparator.
	 * 
	 * @param comparator the dominance comparator
	 */
	public NondominatedSorting(DominanceComparator comparator) {
		super();
		this.comparator = comparator;
	}
	
	/**
	 * Returns the dominance comparator used by this non-dominated sorting
	 * routine.
	 * 
	 * @return the dominance comparator used by this non-dominated sorting
	 *         routine
	 */
	public DominanceComparator getComparator() {
		return comparator;
	}

	/**
	 * Performs non-dominated sorting on the specified population,
	 * assigning the {@code rank} and {@code crowdingDistance} attributes to
	 * solutions.
	 * 
	 * @param population the population whose solutions are to be evaluated
	 */
	public void evaluate(Population population) {
		String algorithm=Executor.algorithmName;
		int n=population.size();		
		int numberOfObjectives=0;
		if(n>0)
		{
	     numberOfObjectives = population.get(0).getNumberOfObjectives();	
		}
		
		if(algorithm=="R-NSGA-II"){ // if the algorithm is R-NSGA-II
		for (int i = 0; i < numberOfObjectives; i++) {
			population.sort(new ObjectiveComparator(i));
			
           //bounds is needed for R-NSGA-II
			bounds[i][0] = population.get(0).getObjective(i);//min for objective i
			bounds[i][1] = population.get(n - 1).getObjective(i); //max for objective i
		}
		}
		
//		if(algorithm=="DF-SMS-EMOA"){
//		int index=0;
//		for(Solution s:population){
//			s.setAttribute(CROWDING_NUMBER, index);
//			index++;
//		}
//		}
		
		List<Solution> remaining = new ArrayList<Solution>();

		for (Solution solution : population) {
			remaining.add(solution);
		}
		
		
				

		int rank = 0;

		while (!remaining.isEmpty()) {
			NondominatedPopulation front = new NondominatedPopulation(
					comparator, DuplicateMode.ALLOW_DUPLICATES);

			for (Solution solution : remaining) {
				front.add(solution);
			}

			for (Solution solution : front) {
				remaining.remove(solution);
				solution.setAttribute(RANK_ATTRIBUTE, rank);
			}

			
			if(algorithm.equalsIgnoreCase("TNSGAII") || algorithm.equalsIgnoreCase("T-NSGA-II")||algorithm.equalsIgnoreCase("T-SMS-EMOA")||algorithm.equalsIgnoreCase("T-R2-EMOA"))
			{
				updateCrowdingDistanceT(front);
			}
			else if(algorithm.equalsIgnoreCase("R-NSGA-II")){
				updateCrowdingDistanceR(front);
			}
			else if(algorithm.equalsIgnoreCase("NSGA-II")||algorithm.equalsIgnoreCase("NSGAII"))
			{
				updateCrowdingDistance(front);
			}
			else
			{
				//algorithm.equalsIgnoreCase("R2-EMOA")||algorithm.equalsIgnoreCase("SMS-EMOA")||algorithm.equalsIgnoreCase("DF-SMS-EMOA")||algorithm.equalsIgnoreCase("AS-EMOA")
				// do nothing, no crowding distance is needed
			
			}


			rank++;
		}
	}

	private void updateCrowdingDistanceR(NondominatedPopulation front) {
		// TODO Auto-generated method stub
		// R-NSGA-II, by longmei
		
		int numberOfTRegion =preference.NumOfROI;
		int numberOfObjectives = front.get(0).getNumberOfObjectives();	
		Double[][] TargetRegions=preference.regions; // only point is considered for R-NSGA-II
		int numberOfSolutionsInFront = front.size();
		
	    Map <Integer, double[]> AttributeMap=new HashMap<Integer, double[]>();// Double[numberOfTRegion] to record EDis to each RP
		
		int count = 0;
		for (Solution solution : front) {
			// set index
			solution.setAttribute(CROWDING_NUMBER, count);
			double[] DisAttribute=new double[numberOfTRegion+1];
			for(int i=0;i<numberOfTRegion;i++)
			{
				double[] RP=new double[numberOfObjectives];
				for(int j=0;j<numberOfObjectives;j++){
					RP[j]=TargetRegions[i][j];
				}
				DisAttribute[i]=CalculatingDistanceToRP(bounds,solution,RP,weights);				
			}
			DisAttribute[numberOfTRegion]=0.0;//0 means have not been checked in epsilon-clearing
			solution.setAttribute(EDistance_ATTRIBUTE, (double)numberOfSolutionsInFront+10);//set initial value to a large number
		  	AttributeMap.put(count, DisAttribute);
	    	count = count+1;	    		    					
		}
//		// epsilon-clearing
//		
		double eps = epsilon;
		for(int t=0;t<numberOfTRegion;t++){
			
			for(Solution solution : front){
				int index= (int) solution.getAttribute(CROWDING_NUMBER);
				double[]AttributeForthisSolution=AttributeMap.get(index);
				if(AttributeForthisSolution!=null)
				solution.setAttribute(SORT_ATTRIBUTE, AttributeForthisSolution[t]);
			}
			
			front.sort(new TchebycheffDistanceComparator1());
			
			for (int i = 0; i < front.size(); i++) {
				double[] AttributeCurrent=AttributeMap.get(front.get(i).getAttribute(CROWDING_NUMBER));
				double LabelCurrent=AttributeCurrent[numberOfTRegion];
				if (LabelCurrent == 0) // has not been checked
				{
					
					for (int j = i+1; j < front.size(); j++) {
						double[] AttributeNext=AttributeMap.get(front.get(j).getAttribute(CROWDING_NUMBER));
						double LabelNext=AttributeNext[numberOfTRegion];
					    boolean InEpsilonNeighourhood=CheckEpsilon(front.get(i),front.get(j),eps);
						// if in the epsilon-neighborhood, change its label and EDis
						if((LabelNext==0)&&InEpsilonNeighourhood)
						{
							AttributeNext[numberOfTRegion]= 1.0;
							double current=(double)front.get(j).getAttribute(SORT_ATTRIBUTE);
							front.get(j).setAttribute(SORT_ATTRIBUTE,  
									(double)front.get(front.size()-1).getAttribute(SORT_ATTRIBUTE)+current);
						}
					}	
				}	
			}
			
			front.sort(new TchebycheffDistanceComparator1());
			// set EDIS attribute
			int numdis = 0;
			for (Solution solution : front) {
				
				if (numdis <= Math.floor(front.size()/numberOfTRegion))// uniform division
				{
					if ((double)solution.getAttribute(EDistance_ATTRIBUTE)>numdis)
					{
						solution.setAttribute(EDistance_ATTRIBUTE, (double)numdis);
//						numdis = numdis + 1;
					}
					numdis = numdis + 1;
				}
				else
				{
					break;
				}
			}
			
			
		}
			
			
		
	}

	private boolean CheckEpsilon(Solution solution1, Solution solution2, double eps) {
		// TODO Auto-generated method stub
		// check whether solution2 is in the epsilon box of solution1, if yes, return true
		boolean result=true;
		int NumOfObj=solution1.getNumberOfObjectives();
		for(int i=0;i<NumOfObj;i++)
		{
			if(Math.abs(solution1.getObjective(i)-solution2.getObjective(i))>eps){
				result=false;
				break;
			}
		}
		
		
		return result;
	}

	//added by Yali for t-nsga-ii
	// improved by Longmei for N numbers of ROI (1<=N)
	public void updateCrowdingDistanceT(Population front) {
		// initially assign all crowding distances of 0.0
		int numberOfTRegion =preference.NumOfROI;
		int numberOfObjectives = front.get(0).getNumberOfObjectives();
		
		List<Double[]> Targets =new ArrayList<Double[]>();
		
		Double[][] TargetRegions=preference.regions; // the array is numberOfTRegion * (numberOfObjectives+1) for circles
		Double[][] CenterPoints= new Double[numberOfTRegion][numberOfObjectives]; //Array of CenterPoints
		
		String Shape=preference.shape;
		
		for(int i=0;i<numberOfTRegion;i++)
		{
			switch(Shape){
			case "circle":
				Double[] R=new Double[numberOfObjectives+1];
				for(int j=0;j<numberOfObjectives+1;j++){
					R[j]=TargetRegions[i][j];
					if(j!=numberOfObjectives)
					{
						CenterPoints[i][j]=TargetRegions[i][j];
					}
				}
				Targets.add(R);
				break;
			case "rectangle":
				Double[] R1=new Double[numberOfObjectives*2];
				for(int j=0;j<numberOfObjectives*2;j++)
					{R1[j]=TargetRegions[i][j];}
				
				Targets.add(R1);

				break;
			case "point":	
				Double[] R2=new Double[numberOfObjectives];
				for(int j=0;j<numberOfObjectives;j++)
				{R2[j]=TargetRegions[i][j];
				CenterPoints[i][j]=TargetRegions[i][j];}
				Targets.add(R2);
				break;
			default:
	             throw new IllegalArgumentException("Invalid shape of preference region: " + Shape);
			}
			
			
		}
		
		// set CenterPoint for rectangles.
		if (Shape=="rectangle")
	{
			for(int i=0;i<numberOfTRegion;i++)
			{
				
				for(int j=0;j<numberOfObjectives;j++)
					
				{
					CenterPoints[i][j]=(TargetRegions[i][j]+TargetRegions[i][j+numberOfObjectives])/2;
				}
				
				
			}
						
				
	}
				
		int numberOfSolutionsInFront = front.size();
		
		// Integer is the INDEX of each solution, Double[] is the array of attributes in the format of 
		//CROWDING1_ATTRIBUTE, CROWDING2_ATTRIBUTE,...CROWDINGM_ATTRIBUTE, M is the number of ROIs
		//Distance1_ATTRIBUTE, Distance2_ATTRIBUTE,...DistanceM_ATTRIBUTE
		//LABEL1_ATTRIBUTE, LABEL2_ATTRIBUTE,...LABELM_ATTRIBUTE
		//
		Map <Integer, Double[][]> AttributeMap=new HashMap<Integer, Double[][]>();
		
		int count = 0;

	
		for (Solution solution : front) {
			// set index
			solution.setAttribute(CROWDING_NUMBER, count);
			
			Double[][]AttributeForSolution=new Double[3][numberOfTRegion];
			
			//initially set crowding distance to 0.
			
			solution.setAttribute(CROWDING_ATTRIBUTE, 0.0);
						
			for(int i=0;i<numberOfTRegion;i++)
			{
				AttributeForSolution[0][i]=0.0;
			}
	       //set AttributeForSolution array (second and third row for Tchebycheff distance and objective label)			
			double smallDis=Double.POSITIVE_INFINITY;
			for(int i=0;i<numberOfTRegion;i++)

			{
				double dis = 0.0;
				int objnum = 0;
				for (int k = 0; k < numberOfObjectives; k++) 
				{
					double Td=Math.abs(solution.getObjective(k)-CenterPoints[i][k]);
					if(Td>dis)
					{
						dis=Td;
						objnum=k;						
					}
					
				}
				AttributeForSolution[1][i]=dis; // Tchebycheff distance to target i
				AttributeForSolution[2][i]=(double)objnum; // objective to get max distance  
				
				if(dis<smallDis)
				{
					smallDis=dis;
				}
			}
			solution.setAttribute(DISTANCE_ATTRIBUTE, numberOfSolutionsInFront+smallDis);
		  	AttributeMap.put(count, AttributeForSolution);
	    	count = count+1;	    		    		
	    	//for test
//	    	Double[][]AttributeForthisSolution=AttributeMap.get("0");
//	    	double k=AttributeForthisSolution[1][0];
		}
		
//		// epsilon-clearing, only works when point is used or no intersection with target region.
	
		double eps = epsilon;
		for(int t=0;t<numberOfTRegion;t++){
			// set AttributeMap[1][t] to SORT_ATTRIBUTE of each solution
			for(Solution solution : front){
				int index= (int) solution.getAttribute(CROWDING_NUMBER);
				Double[][]AttributeForthisSolution=AttributeMap.get(index);
				if(AttributeForthisSolution!=null)
				solution.setAttribute(SORT_ATTRIBUTE, AttributeForthisSolution[1][t]);
			}
			
			front.sort(new TchebycheffDistanceComparator1());
			
			
//			for (int i = 0; i < front.size(); i++) {
//				Double[][] AttributeCurrent=AttributeMap.get(front.get(i).getAttribute(CROWDING_NUMBER));
//				double LabelCurrent=AttributeCurrent[2][t];
//				if (LabelCurrent < numberOfObjectives) // has not been checked
//				{
//					double CurrentDis = (double) front.get(i).getAttribute(SORT_ATTRIBUTE);
//					for (int j = i+1; j < front.size(); j++) {
//						Double[][] AttributeNext=AttributeMap.get(front.get(j).getAttribute(CROWDING_NUMBER));
//						double LabelNext=AttributeNext[2][t];
//						double NextDis = (double) front.get(j).getAttribute(SORT_ATTRIBUTE);
//						// if in the epsilon-neighborhood, change its AttributeMap[1][t] and AttributeMap[2][t]
//						if((LabelNext<numberOfObjectives)&&(LabelNext==LabelCurrent)&&(NextDis<(CurrentDis+eps)))
//						{
//							AttributeNext[2][t]=(double) numberOfObjectives;
//							// for test 
////							Double[][] AttributeTest=AttributeMap.get(front.get(j).getAttribute(CROWDING_NUMBER));
////							double LabelTest=AttributeNext[2][t];
//							// worse than the current worst
//							front.get(j).setAttribute(SORT_ATTRIBUTE,  
//									(double)front.get(front.size()-1).getAttribute(SORT_ATTRIBUTE)+(double)front.get(j).getAttribute(SORT_ATTRIBUTE));
//						}
//					}	
//				}	
//			}
			
			// change epsilon clearing to the method of R-NSGA-II
//			if(Shape=="point"){ //do not use this condition
			
			for (int i = 0; i < front.size(); i++) {
				Double[][] AttributeCurrent=AttributeMap.get(front.get(i).getAttribute(CROWDING_NUMBER));
				double LabelCurrent=AttributeCurrent[2][t];
				if (LabelCurrent < numberOfObjectives) // has not been checked
				{
					
					for (int j = i+1; j < front.size(); j++) {
						Double[][] AttributeNext=AttributeMap.get(front.get(j).getAttribute(CROWDING_NUMBER));
						double LabelNext=AttributeNext[2][t];
					    boolean InEpsilonNeighourhood=CheckEpsilon(front.get(i),front.get(j),eps);
						// if in the epsilon-neighborhood, change its label and EDis
						if((LabelNext< numberOfObjectives)&&InEpsilonNeighourhood)
						{
							AttributeNext[2][t]=(double)numberOfObjectives;
							double current=(double)front.get(j).getAttribute(SORT_ATTRIBUTE);
							front.get(j).setAttribute(SORT_ATTRIBUTE,  
									(double)front.get(front.size()-1).getAttribute(SORT_ATTRIBUTE)+current);
						}
					}	
				}	
			}
			front.sort(new TchebycheffDistanceComparator1());
			
			
//			}// epsilon clearing end
			
			// set Chebycheff Distance attribute	
			int numdis = 0;
			for (Solution solution : front) {
				
				if (numdis <= Math.floor(front.size()/numberOfTRegion))// uniform division
				{
					if ((double)solution.getAttribute(DISTANCE_ATTRIBUTE)>numdis)
					{
						solution.setAttribute(DISTANCE_ATTRIBUTE, (double)numdis);
//						numdis = numdis + 1;
					}
					numdis = numdis + 1;
				}
				else
				{
					break;
				}
			}
			
		}// for number of target regions. End of seting Chebyshev attribute
		
		// start setting crowding_attribute
		String algorithm=Executor.algorithmName;
		if(algorithm.equals("T-NSGA-II")&&Shape!="point"){
			
    //	find solutions in the ROIs
		double[] MaxCrowding= new double[numberOfTRegion];
		int TotalNumberInROI=0;

		for(int t=0;t<numberOfTRegion;t++){
			
			
			Population tempfront = new Population();
			// find solutions in this region
				for(Solution s:front){
					if(Shape=="circle"){
						double distance=0.0;
						double[] centerPoint=new double[numberOfObjectives];
						double radius=TargetRegions[t][numberOfObjectives];
						for (int i=0;i<numberOfObjectives;i++)
						{
							distance+=Math.pow(Math.abs(s.getObjective(i)- CenterPoints[t][i]),2);
							centerPoint[i]=CenterPoints[t][i];
						}
						if (distance <= Math.pow(radius, 2))// in the region
						{
							Solution solutiontemp = coordinatetransformationCT(s, numberOfObjectives, centerPoint, radius);
							tempfront.add(solutiontemp);
						}
					
					}
					else if(Shape=="rectangle"){
						double[] minBound=new double[numberOfObjectives];
						double[] maxBound=new double[numberOfObjectives];
						boolean InROI=true;
						for (int i=0;i<numberOfObjectives;i++)
						{
							minBound[i]=TargetRegions[t][i];
							maxBound[i]=TargetRegions[t][i+numberOfObjectives];
							if((s.getObjective(i)>maxBound[i])||(s.getObjective(i)<minBound[i])){
								InROI=false;
								break;
							}
						}
						if(InROI){
							Solution solutiontemp=coordinatetransformationRT( s,  minBound, maxBound);
							tempfront.add(solutiontemp);
						}
					
				}

				
			}
				Population uniqueFront = new Population(tempfront);
				// remove duplicate solutions
//				Population uniqueFront = new Population();
//				for (Solution s1 : tempfront) {
//					boolean isDuplicate = false;
//					
//					for (Solution s2 : uniqueFront) {
//						//Settings.EPS
//						if (NondominatedPopulation.distance(s1, s2) < Settings.EPS) {
//							isDuplicate = true;
//							break;
//						}
//					}		
//					
//					if (!isDuplicate) {
//						uniqueFront.add(s1);
//					}
//					else
//					{// it seems redundant
//						int tempindex = (int)s1.getAttribute(CROWDING_NUMBER);
//						front.get(tempindex).setAttribute(CROWDING_ATTRIBUTE, 0.0);		
//					}
//				}
				
				// set CrowdingDistance in this region, or AttributeSolution[0][t]
				
				int n = uniqueFront.size();
				TotalNumberInROI+=n;
				if (n < 3&&n>0) {
					for (Solution solution : uniqueFront) {
						int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
						Double[][]AttributeForthisSolution=AttributeMap.get(tempindex);
						AttributeForthisSolution[0][t]=Double.POSITIVE_INFINITY;
						solution.setAttribute(CROWDING_Distance, Double.POSITIVE_INFINITY);
					}
				} else if(n>=3){
					
					for (int i = 0; i < numberOfObjectives; i++) {
						uniqueFront.sort(new ObjectiveComparator(i));
		
						double minObjective = uniqueFront.get(0).getObjective(i);
						double maxObjective = uniqueFront.get(n - 1).getObjective(i);
						// set solutions of ends
						int tempindex = (int)uniqueFront.get(0).getAttribute(CROWDING_NUMBER);
						Double[][]AttributeForthisSolution=AttributeMap.get(tempindex);
						AttributeForthisSolution[0][t]=Double.POSITIVE_INFINITY;
						uniqueFront.get(0).setAttribute(CROWDING_Distance, Double.POSITIVE_INFINITY);
		
						tempindex = (int)uniqueFront.get(n-1).getAttribute(CROWDING_NUMBER);
						AttributeForthisSolution=AttributeMap.get(tempindex);
						AttributeForthisSolution[0][t]=Double.POSITIVE_INFINITY;
						uniqueFront.get(n-1).setAttribute(CROWDING_Distance, Double.POSITIVE_INFINITY);
		// solutions in the middle of this front
						for (int j = 1; j < n - 1; j++) {
							tempindex = (int)uniqueFront.get(j).getAttribute(CROWDING_NUMBER);
							AttributeForthisSolution=AttributeMap.get(tempindex);
							double distance = AttributeForthisSolution[0][t];
							distance += (uniqueFront.get(j + 1).getObjective(i) - 
									uniqueFront.get(j - 1).getObjective(i))
									/ (maxObjective - minObjective);
							AttributeForthisSolution[0][t]=distance;
							uniqueFront.get(j).setAttribute(CROWDING_Distance, distance);
						}
					}
					
				}
			// set CrowdingAttribute, same as the mechanic in distanceAttribute
				if(n>0){
				for(Solution solution : front){
					int index= (int) solution.getAttribute(CROWDING_NUMBER);
					Double[][]AttributeForthisSolution=AttributeMap.get(index);
					if(AttributeForthisSolution!=null)
					solution.setAttribute(SORT_ATTRIBUTE, AttributeForthisSolution[0][t]);
				}
				
				front.sort(new TchebycheffDistanceComparator1());// from 0 to infinity
				
				if(n>=3){
				double maxCrowdingDis = (double)front.get(numberOfSolutionsInFront-3).getAttribute(SORT_ATTRIBUTE);
				MaxCrowding[t]=maxCrowdingDis;
				// for test
//				if(maxCrowdingDis>0)
//				{
//					double d=maxCrowdingDis;
//					
//					System.out.print(d);
//				}
				}
				
				else// if there are less than three solutions, 
				{
					MaxCrowding[t]=1;
				}
		
				int numdis = 0;
					
				for (int m=numberOfSolutionsInFront-1;m>=0;m--) {
						
					if (numdis <= Math.floor(front.size()/numberOfTRegion)) 
					{
						if (((double)front.get(m).getAttribute(CROWDING_ATTRIBUTE)<(numberOfSolutionsInFront-numdis))&& ((double)front.get(m).getAttribute(SORT_ATTRIBUTE)!=0.0))
						
						{
							front.get(m).setAttribute(CROWDING_ATTRIBUTE, (double)(numberOfSolutionsInFront-numdis));
						}
						numdis = numdis + 1;
					}
				}
				}// if n>0 then do the crowding attribute assignment, otherwise keep crowding_attribute as zero
						
		} // for number of targets
				
		// handle solutions without Crowding_ATTRIBUTE, if number of ROI is more than 1 and there are solutions in any target region
			if((numberOfTRegion>1)&&TotalNumberInROI>0){
				
		for (Solution solution : front) {
			double tempdata= (double)solution.getAttribute(CROWDING_ATTRIBUTE);
			int index= (int) solution.getAttribute(CROWDING_NUMBER);
			Double[][]AttributeForthisSolution=AttributeMap.get(index);
			double[] crowdingdata=new double[numberOfTRegion];
			for(int i=0;i<numberOfTRegion;i++){
				crowdingdata[i]=AttributeForthisSolution[0][i];
			}

			if ((tempdata == 0.0) && (HasNoneZeroValue(crowdingdata)))
			{
			 double[] newdata= new double[numberOfTRegion];
				for(int k=0;k<numberOfTRegion;k++){
					double tempdata3 = 0.0;
					if (crowdingdata[k] != 0.0)
					{						
//						tempdata3 = crowdingdata[k];
						// will it be more than 1 ?
						tempdata3 = crowdingdata[k]/numberOfObjectives;
						newdata[k]=tempdata3;
					}
				}
				double value=FindLargestInDoubleArray(newdata);
//				double minvalue=FindSmallestInDoubleArray(newdata);
				solution.setAttribute(CROWDING_ATTRIBUTE, value);	
			}
		} 
		}// if number of ROI more than 1
		}//if(algorithm.equals("T-NSGA-II")||Shape!="point")
		// check front for debug
		int k=front.size();
	}
	
	public void updateCrowdingDistanceTForPrune(Population front) {
		// initially assign all crowding distances of 0.0
		int numberOfTRegion =preference.NumOfROI;
		int numberOfObjectives = front.get(0).getNumberOfObjectives();
		
		List<Double[]> Targets =new ArrayList<Double[]>();
		
		Double[][] TargetRegions=preference.regions; // the array is numberOfTRegion * (numberOfObjectives+1) for circles
		Double[][] CenterPoints= new Double[numberOfTRegion][numberOfObjectives]; //Array of CenterPoints
		
		String Shape=preference.shape;
		
		for(int i=0;i<numberOfTRegion;i++)
		{
			switch(Shape){
			case "circle":
				Double[] R=new Double[numberOfObjectives+1];
				for(int j=0;j<numberOfObjectives+1;j++){
					R[j]=TargetRegions[i][j];
					if(j!=numberOfObjectives)
					{
						CenterPoints[i][j]=TargetRegions[i][j];
					}
				}
				Targets.add(R);
				break;
			case "rectangle":
				Double[] R1=new Double[numberOfObjectives*2];
				for(int j=0;j<numberOfObjectives*2;j++)
					{R1[j]=TargetRegions[i][j];}
				
				Targets.add(R1);

				break;
			case "point":	
				Double[] R2=new Double[numberOfObjectives];
				for(int j=0;j<numberOfObjectives;j++)
				{R2[j]=TargetRegions[i][j];
				CenterPoints[i][j]=TargetRegions[i][j];}
				Targets.add(R2);
				break;
			default:
	             throw new IllegalArgumentException("Invalid shape of preference region: " + Shape);
			}
			
			
		}
		
	
				
		int numberOfSolutionsInFront = front.size();
		
		// Integer is the INDEX of each solution, Double[] is the array of attributes in the format of 
		//CROWDING1_ATTRIBUTE, CROWDING2_ATTRIBUTE,...CROWDINGM_ATTRIBUTE, M is the number of ROIs
		//Distance1_ATTRIBUTE, Distance2_ATTRIBUTE,...DistanceM_ATTRIBUTE
		//LABEL1_ATTRIBUTE, LABEL2_ATTRIBUTE,...LABELM_ATTRIBUTE
		//
		Map <Integer, Double[][]> AttributeMap=new HashMap<Integer, Double[][]>();
		
		int count = 0;

	
		for (Solution solution : front) {
			// set index
			solution.setAttribute(CROWDING_NUMBER, count);
			
			Double[][]AttributeForSolution=new Double[3][numberOfTRegion];
			
			//initially set crowding distance to 0.
			
			solution.setAttribute(CROWDING_ATTRIBUTE, 0.0);
						
			for(int i=0;i<numberOfTRegion;i++)
			{
				AttributeForSolution[0][i]=0.0;
			}
			AttributeMap.put(count, AttributeForSolution);
	  
	    	count = count+1;	    		    		

		}
		

			
    //	find solutions in the ROIs
		double[] MaxCrowding= new double[numberOfTRegion];

		for(int t=0;t<numberOfTRegion;t++){
			
			Population tempfront = new Population();
			// find solutions in this region
				for(Solution s:front){
					if(Shape=="circle"){
						double distance=0.0;
						double[] centerPoint=new double[numberOfObjectives];
						double radius=TargetRegions[t][numberOfObjectives];
						for (int i=0;i<numberOfObjectives;i++)
						{
							distance+=Math.pow(Math.abs(s.getObjective(i)- CenterPoints[t][i]),2);
							centerPoint[i]=CenterPoints[t][i];
						}
						if (distance <= Math.pow(radius, 2))// in the region
						{
							Solution solutiontemp = coordinatetransformationCT(s, numberOfObjectives, centerPoint, radius);
							tempfront.add(solutiontemp);
						}
					
					}
					else if(Shape=="rectangle"){
						double[] minBound=new double[numberOfObjectives];
						double[] maxBound=new double[numberOfObjectives];
						boolean InROI=true;
						for (int i=0;i<numberOfObjectives;i++)
						{
							minBound[i]=TargetRegions[t][i];
							maxBound[i]=TargetRegions[t][i+numberOfObjectives];
							if((s.getObjective(i)>maxBound[i])||(s.getObjective(i)<minBound[i])){
								InROI=false;
							}
						}
						if(InROI){
							Solution solutiontemp=coordinatetransformationRT( s,  minBound, maxBound);
							tempfront.add(solutiontemp);
						}
					
				}

				
			}
				Population uniqueFront = new Population(tempfront);

				
				// set CrowdingDistance in this region, or AttributeSolution[0][t]
				
				int n = uniqueFront.size();
				
				if (n < 3&&n>0) {
					for (Solution solution : uniqueFront) {
						int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
						Double[][]AttributeForthisSolution=AttributeMap.get(tempindex);
						AttributeForthisSolution[0][t]=Double.POSITIVE_INFINITY;
						solution.setAttribute(CROWDING_Distance, Double.POSITIVE_INFINITY);
					}
				} else if(n>=3){
					
					for (int i = 0; i < numberOfObjectives; i++) {
						uniqueFront.sort(new ObjectiveComparator(i));
		
						double minObjective = uniqueFront.get(0).getObjective(i);
						double maxObjective = uniqueFront.get(n - 1).getObjective(i);
						// set solutions of ends
						int tempindex = (int)uniqueFront.get(0).getAttribute(CROWDING_NUMBER);
						Double[][]AttributeForthisSolution=AttributeMap.get(tempindex);
						AttributeForthisSolution[0][t]=Double.POSITIVE_INFINITY;
						uniqueFront.get(0).setAttribute(CROWDING_Distance, Double.POSITIVE_INFINITY);
		
						tempindex = (int)uniqueFront.get(n-1).getAttribute(CROWDING_NUMBER);
						AttributeForthisSolution=AttributeMap.get(tempindex);
						AttributeForthisSolution[0][t]=Double.POSITIVE_INFINITY;
						uniqueFront.get(n-1).setAttribute(CROWDING_Distance, Double.POSITIVE_INFINITY);
		// solutions in the middle of this front
						for (int j = 1; j < n - 1; j++) {
							tempindex = (int)uniqueFront.get(j).getAttribute(CROWDING_NUMBER);
							AttributeForthisSolution=AttributeMap.get(tempindex);
							double distance = AttributeForthisSolution[0][t];
							distance += (uniqueFront.get(j + 1).getObjective(i) - 
									uniqueFront.get(j - 1).getObjective(i))
									/ (maxObjective - minObjective);
							AttributeForthisSolution[0][t]=distance;
							uniqueFront.get(j).setAttribute(CROWDING_Distance, distance);
						}
					}
					
				}
			// set CrowdingAttribute, same as the mechanic in distanceAttribute
				if(n>0){
				for(Solution solution : front){
					int index= (int) solution.getAttribute(CROWDING_NUMBER);
					Double[][]AttributeForthisSolution=AttributeMap.get(index);
					if(AttributeForthisSolution!=null)
					solution.setAttribute(SORT_ATTRIBUTE, AttributeForthisSolution[0][t]);
				}
				
				front.sort(new TchebycheffDistanceComparator1());// from 0 to infinity
				if(numberOfSolutionsInFront>=3){
				double maxCrowdingDis = (double)front.get(numberOfSolutionsInFront-3).getAttribute(SORT_ATTRIBUTE);
				MaxCrowding[t]=maxCrowdingDis;
				// for test
//				if(maxCrowdingDis>0)
//				{
//					double d=maxCrowdingDis;
//					
//					System.out.print(d);
//				}
				}
				
				else// if there are less than three solutions, 
				{
					MaxCrowding[t]=1;
				}
		
				int numdis = 0;
					
				for (int m=numberOfSolutionsInFront-1;m>=0;m--) {
						
					if (numdis <= Math.floor(front.size()/numberOfTRegion)) 
					{
						if (((double)front.get(m).getAttribute(CROWDING_ATTRIBUTE)<(numberOfSolutionsInFront-numdis))&& ((double)front.get(m).getAttribute(SORT_ATTRIBUTE)!=0.0))
						
						{
							front.get(m).setAttribute(CROWDING_ATTRIBUTE, (double)(numberOfSolutionsInFront-numdis));
						}
						numdis = numdis + 1;
					}
				}
						
		} // for number of targets
				
		// handle solutions without Crowding_ATTRIBUTE, if number of ROI is more than 1
			if(numberOfTRegion>1){
				
		for (Solution solution : front) {
			double tempdata= (double)solution.getAttribute(CROWDING_ATTRIBUTE);
			int index= (int) solution.getAttribute(CROWDING_NUMBER);
			Double[][]AttributeForthisSolution=AttributeMap.get(index);
			double[] crowdingdata=new double[numberOfTRegion];
			for(int i=0;i<numberOfTRegion;i++){
				crowdingdata[i]=AttributeForthisSolution[0][i];
			}

			if ((tempdata == 0.0) && (HasNoneZeroValue(crowdingdata)))
			{
			 double[] newdata= new double[numberOfTRegion];
				for(int k=0;k<numberOfTRegion;k++){
					double tempdata3 = 0.0;
					if (crowdingdata[k] != 0.0)
					{						
//						tempdata3 = crowdingdata[k];
						// will it be more than 1 ?
						tempdata3 = crowdingdata[k]/numberOfObjectives;
						newdata[k]=tempdata3;
					}
				}
				double value=FindLargestInDoubleArray(newdata);
//				double minvalue=FindSmallestInDoubleArray(newdata);
				solution.setAttribute(CROWDING_ATTRIBUTE, value);	
			}
		} 
		}// if number of ROI more than 1
		}// if population size in ROI is more than 1
}
	
	
	private double FindSmallestInDoubleArray(double[] newdata) {
		 // Validates input
	     if (newdata == null) {
	         throw new IllegalArgumentException("The Array must not be null");
	     } else if (newdata.length == 0) {
	         throw new IllegalArgumentException("Array cannot be empty.");
	     }
	 
	     // Finds and returns min
	     double min = newdata[0];
	     for (int i = 1; i < newdata.length; i++) {
	         if (Double.isNaN(newdata[i])) {
	             return Double.NaN;
	         }
	         if (newdata[i] < min) {
	             min = newdata[i];
	         }
	     }
	 
	     return min;
	}

	private double FindLargestInDoubleArray(double[] newdata) {
		 // Validates input
	     if (newdata == null) {
	         throw new IllegalArgumentException("The Array must not be null");
	     } else if (newdata.length == 0) {
	         throw new IllegalArgumentException("Array cannot be empty.");
	     }
	 
	     // Finds and returns max
	     double max = newdata[0];
	     for (int i = 1; i < newdata.length; i++) {
	         if (Double.isNaN(newdata[i])) {
	             return Double.NaN;
	         }
	         if (newdata[i] > max) {
	        	 max = newdata[i];
	         }
	     }
	 
	     return max;
	}

	
	private boolean HasNoneZeroValue(double[] crowdingdata) {
		// TODO at least one of the elements in crowdingdata is not 0
		boolean redult=false;
		int num=crowdingdata.length;
		for(int i=0;i<num;i++){
			if(crowdingdata[i]!=0.0)
			{
				redult=true;
				break;
			}
		}
		return redult;
	}

	public static Solution coordinatetransformationCT(Solution solution,int numberOfObjectives, double[] centerPoint, double radius) {
		
		Solution newSolution = solution.copy();
		
		int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
		newSolution.setAttribute(CROWDING_NUMBER, tempindex);
		double tempdis = (double)solution.getAttribute(CROWDING_ATTRIBUTE);
		newSolution.setAttribute(CROWDING_ATTRIBUTE, tempdis);
		double sort = (double)solution.getAttribute(SORT_ATTRIBUTE);
		newSolution.setAttribute(SORT_ATTRIBUTE, sort);
		
		double tempChdis = (double)solution.getAttribute(DISTANCE_ATTRIBUTE);
		newSolution.setAttribute(DISTANCE_ATTRIBUTE, tempChdis);
//	    tempdis = (double)solution.getAttribute(CROWDING1_ATTRIBUTE);
//		newSolution.setAttribute(CROWDING1_ATTRIBUTE, tempdis);		
//		tempdis = (double)solution.getAttribute(CROWDING2_ATTRIBUTE);
//		newSolution.setAttribute(CROWDING2_ATTRIBUTE, tempdis);		
		int rank=(int) solution.getAttribute(RANK_ATTRIBUTE);
		newSolution.setAttribute(RANK_ATTRIBUTE, rank);
		
		for (int i=0;i<numberOfObjectives;i++)
		{
			newSolution.setObjective(i, newSolution.getObjective(i)-(centerPoint[i]-radius));	
		}	
					
		return newSolution;
	}
	
	
	public static Solution coordinatetransformationRT(Solution solution,double[] minBound, double[] maxBound) {
		
		Solution newSolution = solution.copy();
		
		
			int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
			newSolution.setAttribute(CROWDING_NUMBER, tempindex);
			double tempdis = (double)solution.getAttribute(DISTANCE_ATTRIBUTE);
			newSolution.setAttribute(DISTANCE_ATTRIBUTE, tempdis);
			double tempCdis = (double)solution.getAttribute(CROWDING_ATTRIBUTE);
			newSolution.setAttribute(CROWDING_ATTRIBUTE, tempCdis);
			int rank= (int) solution.getAttribute(RANK_ATTRIBUTE);
			newSolution.setAttribute(RANK_ATTRIBUTE, rank);
			double sort = (double)solution.getAttribute(SORT_ATTRIBUTE);
			newSolution.setAttribute(SORT_ATTRIBUTE, sort);
			
			for (int i = 0; i < solution.getNumberOfObjectives(); i++)
			{
				newSolution.setObjective(i, TRFunction(solution.getObjective(i), i, minBound, maxBound));
			}
					
		return newSolution;
	}
	
	public static List<Solution> coordinatetransformation(Population population,double[] minBound, double[] maxBound) {
		
		List<Solution> result = new ArrayList<Solution>();
		double tempdis = 0.0;
		
		for (Solution solution : population) {
			Solution newSolution = solution.copy();
			
			int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
			newSolution.setAttribute(CROWDING_NUMBER, tempindex);
			tempdis = (double)solution.getAttribute(DISTANCE_ATTRIBUTE);
			newSolution.setAttribute(DISTANCE_ATTRIBUTE, tempdis);
			tempdis = (double)solution.getAttribute(CROWDING_ATTRIBUTE);
			newSolution.setAttribute(CROWDING_ATTRIBUTE, tempdis);
			int rank= (int) solution.getAttribute(RANK_ATTRIBUTE);
			newSolution.setAttribute(RANK_ATTRIBUTE, rank);
			
			for (int i = 0; i < solution.getNumberOfObjectives(); i++)
			{
				newSolution.setObjective(i, TRFunction(solution.getObjective(i), i, minBound, maxBound));
			}
			
			
			result.add(newSolution);
		}
		
		return result;
	}
	
	public static double TRFunction(double y, int objective, double[] minBound, double[] maxBound){
		double result=0;
		
		if (y <= minBound[objective])
		{
			result = 0;
		}
		else if(y < maxBound[objective])
		{
			result = y - minBound[objective];
		}
		else
		{
			result = maxBound[objective] - minBound[objective];
		}
	
		return result;
	}
	
	public static double TRFunctionReverse(double y, int objective, double[] minBound, double[] maxBound){
		double result=0;
		
		if (y >= maxBound[objective])
		{
			result = 0;
		}
		else if(y > minBound[objective])
		{
			result = maxBound[objective]-y;
		}
		else
		{
			result = maxBound[objective] - minBound[objective];
		}
	
		return result;
	}
	
	//calculate distance to the reference point.
		//bounds:max and min values for each objective in the population
		//s: the solution to be calculated
		//RP:reference point
		//weights: weight vector, same dimension as RP
		public double CalculatingDistanceToRP(double[][] bounds,Solution s,double[]RP,double[]weights){
			double dis=0;
			int numberOfObjectives = s.getNumberOfObjectives();
			// array of max and min for each objective
			
				
				for (int k = 0; k < numberOfObjectives; k++) {
					double d=Math.pow((s.getObjective(k)-RP[k])/(bounds[k][1]-bounds[k][0]),2);
			//		double d=Math.pow((s.getObjective(k)-RP[k]),2);	
					dis+=weights[k]*d;
				
				}
				dis=Math.sqrt(dis);	
			
			return dis;
			
		}
		
	/**
	 * Computes and assigns the {@code crowdingDistance} attribute to solutions.
	 * The specified population should consist of solutions within the same
	 * front/rank.
	 * 
	 * @param front the population whose solutions are to be evaluated
	 */
	public void updateCrowdingDistance(Population front) {
		// initially assign all crowding distances of 0.0
		for (Solution solution : front) {
			solution.setAttribute(CROWDING_ATTRIBUTE, 0.0);
		}
		
		// remove any duplicate solutions, the duplicate solutions will retain
		// the crowding distance of 0.0
		Population uniqueFront = new Population();
		
		for (Solution s1 : front) {
			boolean isDuplicate = false;
			
			for (Solution s2 : uniqueFront) {
				if (NondominatedPopulation.distance(s1, s2) < Settings.EPS) {
					isDuplicate = true;
					break;
				}
			}
			
			if (!isDuplicate) {
				uniqueFront.add(s1);
			}
		}
		
		front = uniqueFront;

		// then compute the crowding distance for the unique solutions
		int n = front.size();
		
		if (n < 3) {
			for (Solution solution : front) {
				solution.setAttribute(CROWDING_ATTRIBUTE,
						Double.POSITIVE_INFINITY);
			}
		} else {
			int numberOfObjectives = front.get(0).getNumberOfObjectives();

			for (int i = 0; i < numberOfObjectives; i++) {
				front.sort(new ObjectiveComparator(i));

				double minObjective = front.get(0).getObjective(i);
				double maxObjective = front.get(n - 1).getObjective(i);

				front.get(0).setAttribute(CROWDING_ATTRIBUTE,
						Double.POSITIVE_INFINITY);
				front.get(n - 1).setAttribute(CROWDING_ATTRIBUTE,
						Double.POSITIVE_INFINITY);

				for (int j = 1; j < n - 1; j++) {
					double distance = (Double)front.get(j).getAttribute(
							CROWDING_ATTRIBUTE);
					distance += (front.get(j + 1).getObjective(i) - 
							front.get(j - 1).getObjective(i))
							/ (maxObjective - minObjective);
					front.get(j).setAttribute(CROWDING_ATTRIBUTE, distance);
				}
			}
		}
	}

}
