/* Copyright 2009-2016 David Hadka
 *
 * This file is part of the MOEA Framework.
 *
 * The MOEA Framework is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * The MOEA Framework is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the MOEA Framework.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.moeaframework.core.fitness;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.moeaframework.algorithm.AbstractAlgorithm;
import org.moeaframework.core.FitnessEvaluator;
import org.moeaframework.core.Population;
import org.moeaframework.core.Problem;
import org.moeaframework.core.Settings;
import org.moeaframework.core.Solution;
import org.moeaframework.core.NondominatedSorting.preference;
import org.moeaframework.core.comparator.FitnessComparator1;
import org.moeaframework.core.comparator.FitnessComparator2;
import org.moeaframework.core.comparator.ObjectiveComparator;
import org.moeaframework.core.comparator.TchebycheffDistanceComparator1;
import org.moeaframework.core.comparator.TchebycheffDistanceComparator2;
import org.moeaframework.core.indicator.Hypervolume;


/**
 * Assigns the fitness of solutions based on their contribution to the overall
 * hypervolume.
 */
public class HypervolumeContributionFitnessEvaluator implements FitnessEvaluator {
	
	/**
	 * The problem.
	 */
	private final Problem problem;
	
	/**
	 * The offset used when determining the reference point for the hypervolume
	 * calculation.
	 */
	private final double offset;
	public static final String RANK_ATTRIBUTE = "rank";

	/**
	 * Attribute key for the crowding distance of a solution.
	 */
	public static final String CROWDING_ATTRIBUTE = "crowdingDistance";
	
	// convert to double arrays
//	public static final String CROWDING1_ATTRIBUTE = "crowdingDistance1";
//	public static final String CROWDING2_ATTRIBUTE = "crowdingDistance2";
	
	public static final String CROWDING_NUMBER = "INDEX";
	
	public static final String DISTANCE_ATTRIBUTE = "TchebycheffDistance";
	public static final String SORT_ATTRIBUTE = "SortingCriterion"; 
	
//	// For DF-SMS-EMOA
//	private static double[] usl={0.3,1}; // upper bound of the region
//	private static double[] lsl={0,0.6}; // lower bound of the region
	// control the shape of Harrington's two-sided DF, the higher, the more it looks like step function
	private int n=5;
	private double epsilon=1.1e-10;
	
	/**
	 * Constructs a new hypervolume contribution fitness evaluator with an
	 * offset of 100.
	 * 
	 * @param problem the problem
	 */
	public HypervolumeContributionFitnessEvaluator(Problem problem) {
		// change offset to smaller value
		this(problem, 100.0);
	}
	
	/**
	 * Constructs a new hypervolume contribution fitness evaluator.
	 * 
	 * @param problem the problem
	 * @param offset the offset used when determining the reference point for\
	 *               the hypervolume calculation.
	 */
	public HypervolumeContributionFitnessEvaluator(Problem problem, double offset) {
		super();
		this.problem = problem;
		this.offset = offset;
	}

	@Override
	public void evaluate(Population population) {

 	 
		if (population.size() <= 2) {
			for (Solution solution : population) {
				solution.setAttribute(FITNESS_ATTRIBUTE, 0.0);
			}
		} else {
			int numberOfObjectives = problem.getNumberOfObjectives();
			List<Solution> solutions = normalize(population);
			List<Solution> solutionsCopy = new ArrayList<Solution>(solutions);
			
			double totalVolume = Hypervolume.calculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives);
			
			for (int i = 0; i < population.size(); i++) {
				solutionsCopy = new ArrayList<Solution>(solutions);
				solutionsCopy.remove(i);
				
				double volume = Hypervolume.calculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives);
				population.get(i).setAttribute(FITNESS_ATTRIBUTE, totalVolume - volume);
			}
		}
	}
	
	/**
	 * Normalizes the population using a reference point calculated by the
	 * maximum extent of the population plus an offset.
	 * 
	 * @param population the population to normalize
	 * @return the normalized solutions
	 */
	private List<Solution> normalizeDF(Population population) {
		List<Solution> result = new ArrayList<Solution>();
		
		double[] min = new double[problem.getNumberOfObjectives()];
		double[] max = new double[problem.getNumberOfObjectives()];
		
		Arrays.fill(min, Double.POSITIVE_INFINITY);
		Arrays.fill(max, Double.NEGATIVE_INFINITY);
		
		for (Solution solution : population) {
			for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
				min[i] = Math.min(min[i], solution.getObjective(i));
//				max[i] = Math.max(max[i], solution.getObjective(i));
				max[i]=0;
			}		
		}
		
		for (Solution solution : population) {
			Solution newSolution = solution.copy();
			
			for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
//				if((max[i] - min[i])!=0)
//				{
					//newSolution.setObjective(i, (max[i] - (newSolution.getObjective(i) - min[i]) + offset) / (max[i] - min[i]));
					// no normalization, use 0 as reference point, only change 
					newSolution.setObjective(i, max[i] - newSolution.getObjective(i));
//				}
//				else{
//				//	newSolution.setObjective(i, max[i] - (newSolution.getObjective(i) - min[i]));
//				}
			}

			result.add(newSolution);
		}
		
		return result;
	}
	
	private List<Solution> normalize(Population population) {
		List<Solution> result = new ArrayList<Solution>();
		
		double[] min = new double[problem.getNumberOfObjectives()];
		double[] max = new double[problem.getNumberOfObjectives()];
		
		Arrays.fill(min, Double.POSITIVE_INFINITY);
		Arrays.fill(max, Double.NEGATIVE_INFINITY);
		
		for (Solution solution : population) {
			for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
				min[i] = Math.min(min[i], solution.getObjective(i));
				max[i] = Math.max(max[i], solution.getObjective(i));
//				max[i]=0;
			}		
		}
		
		for (Solution solution : population) {
			Solution newSolution = solution.copy();
			
			for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
				if((max[i] - min[i])!=0)
				{
					newSolution.setObjective(i, (max[i] - (newSolution.getObjective(i) - min[i]) + offset) / (max[i] - min[i]));
					// no normalization, use 0 as reference point, only change 
//					newSolution.setObjective(i, max[i] - newSolution.getObjective(i));
				}
				else{
					newSolution.setObjective(i, max[i] - (newSolution.getObjective(i) - min[i]));
				}
			}

			result.add(newSolution);
		}
		
		return result;
	}
	
// impoortant normalization!	
	private List<Solution> normalize(List<Solution> solutions) {
		
	List<Solution> result = new ArrayList<Solution>();
		
		double[] min = new double[problem.getNumberOfObjectives()];
		double[] max = new double[problem.getNumberOfObjectives()];
		
		Arrays.fill(min, Double.POSITIVE_INFINITY);
		Arrays.fill(max, Double.NEGATIVE_INFINITY);
		
		for (Solution solution : solutions) {
			for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
				min[i] = Math.min(min[i], solution.getObjective(i));
				max[i] = Math.max(max[i], solution.getObjective(i));
			}
		}
		
		for (Solution solution : solutions) {
			Solution newSolution = solution.copy();
			
			if(solution.getAttribute(RANK_ATTRIBUTE)!=null)
			{
				newSolution.setAttribute(RANK_ATTRIBUTE, (int)solution.getAttribute(RANK_ATTRIBUTE));
			}
			
			if(solution.getAttribute(CROWDING_NUMBER)!=null)
			{
				newSolution.setAttribute(CROWDING_NUMBER, (int)solution.getAttribute(CROWDING_NUMBER));
			}
			
			for (int i = 0; i < problem.getNumberOfObjectives(); i++) {
				if((max[i] - min[i])!=0){
				newSolution.setObjective(i, (max[i] - (newSolution.getObjective(i) - min[i]) + offset) / (max[i] - min[i]));}
			//	newSolution.setObjective(i, max[i] - (newSolution.getObjective(i) - min[i]));}
				else// only one solution
				{
					newSolution.setObjective(i, 1 - newSolution.getObjective(i));
				}
			}

			result.add(newSolution);
		}
		
		return result;
	}
	
	public void evaluateT(Population population) {
		int numberOfSolutionsInFront=population.size();
		int numberOfTRegion =preference.NumOfROI;
		String Shape=preference.shape;
		int numberOfObjectives = population.get(0).getNumberOfObjectives();
		boolean checkSolutionInROIwithoutattribute=false;
		if(Shape.equalsIgnoreCase("point")){
			for(Solution s:population)
			{
				s.setAttribute(FITNESS_ATTRIBUTE, 0.0);
			}
		} // if reference point is used, no need to calculate fitness attribute
		else{
		List<Double[]> Targets =new ArrayList<Double[]>();		
		Double[][] TargetRegions=preference.regions; // the array is numberOfTRegion * (numberOfObjectives+1) for circles
		Double[][] CenterPoints= new Double[numberOfTRegion][numberOfObjectives]; //Array of CenterPoints
		 Map <Integer, double[]> AttributeMap=new HashMap<Integer, double[]>();// double[numberOfTRegion] to record delta-HV in each ROI
	
		
	//	AttributeMap	initialization
		for(Solution s:population)
		{
			double[] HVAttribute=new double[numberOfTRegion];
			for(int t=0;t<numberOfTRegion;t++){
				HVAttribute[t]=0;
			}
			AttributeMap.put((int)s.getAttribute(CROWDING_NUMBER), HVAttribute);
			s.setAttribute(FITNESS_ATTRIBUTE, 0.0);
		}
	// CenterPoints	
		for(int i=0;i<numberOfTRegion;i++)
		{
			switch(Shape){
			case "circle":
				Double[] R=new Double[numberOfObjectives+1];
				for(int j=0;j<numberOfObjectives+1;j++){
					R[j]=TargetRegions[i][j];
					if(j!=numberOfObjectives)
					{
						CenterPoints[i][j]=TargetRegions[i][j];
					}
				}
				Targets.add(R);
				break;
			case "rectangle":
				Double[] R1=new Double[numberOfObjectives*2];
				for(int j=0;j<numberOfObjectives*2;j++)
					{R1[j]=TargetRegions[i][j];}
				
				Targets.add(R1);

				break;
			case "point":	
				Double[] R2=new Double[numberOfObjectives];
				for(int j=0;j<numberOfObjectives;j++)
				{R2[j]=TargetRegions[i][j];
				CenterPoints[i][j]=TargetRegions[i][j];}
				Targets.add(R2);
				break;
			default:
	             throw new IllegalArgumentException("Invalid shape of preference region: " + Shape);
			}
			
			
		}
		
		// set CenterPoint for rectangles.
		if (Shape=="rectangle")
	{
			for(int i=0;i<numberOfTRegion;i++)
			{
				
				for(int j=0;j<numberOfObjectives;j++)
					
				{
					CenterPoints[i][j]=(TargetRegions[i][j]+TargetRegions[i][j+numberOfObjectives])/2;
				}
				
				
			}
						
				
	}

			double[] MaxHVContribution= new double[numberOfTRegion];
//			find solutions in the ROIs
			for(int t=0;t<numberOfTRegion;t++){
				
				List<Solution> tempfront = new ArrayList<Solution>();
				// find solutions in this region
					for(Solution s:population){
						if(Shape=="circle"){
							double distance=0.0;
							double[] centerPoint=new double[numberOfObjectives];
							double radius=TargetRegions[t][numberOfObjectives];
							for (int i=0;i<numberOfObjectives;i++)
							{
								distance+=Math.pow(Math.abs(s.getObjective(i)- CenterPoints[t][i]),2);
								centerPoint[i]=CenterPoints[t][i];
							}
							if (distance <= Math.pow(radius, 2))// in the region
							{
								Solution solutiontemp = coordinatetransformationCTSMS(s, numberOfObjectives, centerPoint, radius);
								tempfront.add(solutiontemp);
							}
						
						}
						else if(Shape=="rectangle"){
							double[] minBound=new double[numberOfObjectives];
							double[] maxBound=new double[numberOfObjectives];
							boolean InROI=true;
							for (int i=0;i<numberOfObjectives;i++)
							{
								minBound[i]=TargetRegions[t][i];
								maxBound[i]=TargetRegions[t][i+numberOfObjectives];
								if((s.getObjective(i)>maxBound[i])||(s.getObjective(i)<minBound[i])){
									InROI=false;
									break;
								}
							}
							if(InROI){
							//	Solution solutiontemp=coordinatetransformationRT( s,  minBound, maxBound);
								// try using maximization Coordinate transformation without normalization HV
								Solution solutiontemp=coordinatetransformationRTSMS( s,  minBound, maxBound);
								tempfront.add(solutiontemp);
							}
						
					}
										
				}
					
         if(tempfront.size()>0){// there exist solutions in ROI
	        checkSolutionInROIwithoutattribute=true;
	        // try no normalization
	//		List<Solution> solutions = normalize(tempfront);
	        List<Solution> solutions = tempfront;
//			List<Solution> solutionsCopy = normalize(solutions);
			List<Solution> solutionsCopy = new ArrayList<Solution>(solutions);

			//double totalVolume = Hypervolume.NewcalculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives, minBound, maxBound, referencePoint);
			double totalVolume = Hypervolume.calculateHypervolume(solutionsCopy, solutionsCopy.size(),numberOfObjectives);

			if (tempfront.size()==1)
			{
				int index=(int) tempfront.get(0).getAttribute(INDEX_ATTRIBUTE);
				population.get(index).setAttribute(FITNESS_ATTRIBUTE, (double)totalVolume);
//			
//				int tempindex = (int) tempfront.get(0).getAttribute(INDEX_ATTRIBUTE);
//				double[] HVAtemp=AttributeMap.get(tempindex);
//				HVAtemp[t]=totalVolume;
			}
			else if(tempfront.size()>1)
			{
				for (int i = 0; i < tempfront.size(); i++) {
					solutionsCopy = new ArrayList<Solution>(solutions);
					solutionsCopy.remove(i);
					
				    //double volume = Hypervolume.NewcalculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives,minBound, maxBound, referencePoint);
					double volume = Hypervolume.calculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives);
					double HVContribution=totalVolume - volume;
					int tempindex = (int) tempfront.get(i).getAttribute(INDEX_ATTRIBUTE);
					double[] HVAtemp=AttributeMap.get(tempindex);
					HVAtemp[t]=HVContribution;
				}	
			
			
			// set Fitness
			for(Solution solution : population){
				int index= (int) solution.getAttribute(CROWDING_NUMBER);
				double[] AttributeForthisSolution=AttributeMap.get(index);
				if(AttributeForthisSolution!=null)
				solution.setAttribute(SORT_ATTRIBUTE, AttributeForthisSolution[t]);
			}
			
			population.sort(new TchebycheffDistanceComparator1());// from 0 to infinity
			
			double maxHVC = (double)population.get(numberOfSolutionsInFront-1).getAttribute(SORT_ATTRIBUTE);
			MaxHVContribution[t]=maxHVC;
			
			int numdis = 0;
				
			for (int m=numberOfSolutionsInFront-1;m>=0;m--) {
					
				if (numdis <= Math.floor(numberOfSolutionsInFront/numberOfTRegion)) 
				{
					if (((double)population.get(m).getAttribute(FITNESS_ATTRIBUTE)<(double)(numberOfSolutionsInFront-numdis))&& ((double)population.get(m).getAttribute(SORT_ATTRIBUTE)!=0.0))
					
					{
						population.get(m).setAttribute(FITNESS_ATTRIBUTE, (double)(numberOfSolutionsInFront+1-numdis));
					}
					numdis = numdis + 1;
				}
				else
				{
					break;
				}
			}
			}// number of solutions in the box bigger than 1
}// number of solutions in the box bigger than 0
	
	} // for number of targets
			if((numberOfTRegion>1)&&(checkSolutionInROIwithoutattribute)){
			// handle solutions without Fitness_ATTRIBUTE
			for (Solution solution : population) {
				double tempdata= (double)solution.getAttribute(FITNESS_ATTRIBUTE);
				int index= (int) solution.getAttribute(CROWDING_NUMBER);
				double[] AttributeForthisSolution=AttributeMap.get(index);
				double[] crowdingdata=new double[numberOfTRegion];
				for(int i=0;i<numberOfTRegion;i++){
					crowdingdata[i]=AttributeForthisSolution[i];
				}

				if ((tempdata == 0.0) && (HasNoneZeroValue(crowdingdata)))
				{
				 double[] newdata= new double[numberOfTRegion];
					for(int k=0;k<numberOfTRegion;k++){
						double tempdata3 = 0.0;
						if ((crowdingdata[k] != 0.0) && (MaxHVContribution[k] !=0.0))
						{
							
						//	tempdata3 = crowdingdata[k];
							// will it be more than 1 ?
							tempdata3 = crowdingdata[k]/MaxHVContribution[k];
							newdata[k]=tempdata3;
						}
					}
					double value=FindLargestInDoubleArray(newdata);
//					double minvalue=FindSmallestInDoubleArray(newdata);
					solution.setAttribute(FITNESS_ATTRIBUTE, value);	
				}
			}
			}// if number of ROI is more than 2
		}// if the shape is not point
		}
	
	


	public static Solution coordinatetransformationCT(Solution solution,int numberOfObjectives, double[] centerPoint, double radius) {
		
		Solution newSolution = solution.copy();
		boolean isInverted = true;
		boolean isCustomHypervolume = (Settings.getHypervolume() != null) && (numberOfObjectives > 2) ;
		
		if (isCustomHypervolume) {
			isInverted = Settings.isHypervolumeInverted();
		}
		
		int tempindex = (int)solution.getAttribute(INDEX_ATTRIBUTE);
		newSolution.setAttribute(INDEX_ATTRIBUTE, tempindex);

		if (isInverted)
		{
			for (int i=0;i<numberOfObjectives;i++)
			{
				newSolution.setObjective(i, newSolution.getObjective(i)-(centerPoint[i]-radius));	
			}	
		}
		else
		{
			for (int i=0;i<numberOfObjectives;i++)
			{
				
				newSolution.setObjective(i, 2*centerPoint[i]-newSolution.getObjective(i)-(centerPoint[i]-radius));		
				
			}
		}
	
				
		return newSolution;
	}
	
	
	public static Solution coordinatetransformationCTSMS(Solution solution,int numberOfObjectives, double[] centerPoint, double radius) {
		
		Solution newSolution = solution.copy();
		boolean isInverted = true;
		boolean isCustomHypervolume = (Settings.getHypervolume() != null) && (numberOfObjectives > 2) ;
		
		if (isCustomHypervolume) {
			isInverted = Settings.isHypervolumeInverted();
		}
		
		int tempindex = (int)solution.getAttribute(INDEX_ATTRIBUTE);
		newSolution.setAttribute(INDEX_ATTRIBUTE, tempindex);

		if (!isInverted)
		{
			for (int i=0;i<numberOfObjectives;i++)
			{
				newSolution.setObjective(i, newSolution.getObjective(i)-(centerPoint[i]-radius));	
			}	
		}
		else
		{
			for (int i=0;i<numberOfObjectives;i++)
			{
				
				newSolution.setObjective(i, 2*centerPoint[i]-newSolution.getObjective(i)-(centerPoint[i]-radius));		
				
			}
		}
	
				
		return newSolution;
	}
	
	
	public static List<Solution> coordinatetransformation(Population population,double[] minBound, double[] maxBound) {
		
		List<Solution> result = new ArrayList<Solution>();
		boolean isInverted = true;
		boolean isCustomHypervolume = (Settings.getHypervolume() != null) && 
				(population.get(0).getNumberOfObjectives() > 2) ;
		
		if (isCustomHypervolume) {
			isInverted = Settings.isHypervolumeInverted();
		}

		
		for (Solution solution : population) {
			Solution newSolution = solution.copy();
			int tempindex = (int)solution.getAttribute(INDEX_ATTRIBUTE);
			newSolution.setAttribute(INDEX_ATTRIBUTE, tempindex);
			
			for (int i = 0; i < solution.getNumberOfObjectives(); i++)
			{
				if (isInverted)
				{
					newSolution.setObjective(i, TRFunctionReverse(solution.getObjective(i), i, minBound, maxBound));
				}
				else
				{
					newSolution.setObjective(i, TRFunction(solution.getObjective(i), i, minBound, maxBound));				
				}
			}
			result.add(newSolution);
		}
		
		return result;
	}
	
	public static double TRFunction(double y, int objective, double[] minBound, double[] maxBound){
		double result=0;
		
		if (y <= minBound[objective])
		{
			result = 0;
		}
		else if(y < maxBound[objective])
		{
			result = y - minBound[objective];
		}
		else
		{
			result = maxBound[objective] - minBound[objective];
		}
	
		return result;
	}
	
	public static double TRFunctionReverse(double y, int objective, double[] minBound, double[] maxBound){
		double result=0;
		
		if (y >= maxBound[objective])
		{
			result = 0;
		}
		else if(y > minBound[objective])
		{
			result = maxBound[objective]-y;
		}
		else
		{
			result = maxBound[objective] - minBound[objective];
		}
	
		return result;
	}
	
public static Solution coordinatetransformationRT(Solution solution,double[] minBound, double[] maxBound) {
		
		Solution newSolution = solution.copy();
		
		
			int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
			newSolution.setAttribute(CROWDING_NUMBER, tempindex);
			double tempdis = (double)solution.getAttribute(DISTANCE_ATTRIBUTE);
			newSolution.setAttribute(DISTANCE_ATTRIBUTE, tempdis);
			double tempCdis = (double)solution.getAttribute(CROWDING_ATTRIBUTE);
			newSolution.setAttribute(CROWDING_ATTRIBUTE, tempCdis);
			int rank= (int) solution.getAttribute(RANK_ATTRIBUTE);
			newSolution.setAttribute(RANK_ATTRIBUTE, rank);
			
			for (int i = 0; i < solution.getNumberOfObjectives(); i++)
			{
				newSolution.setObjective(i, TRFunction(solution.getObjective(i), i, minBound, maxBound));
			}
					
		return newSolution;
	}

public static Solution coordinatetransformationRTSMS(Solution solution,double[] minBound, double[] maxBound) {
	
	Solution newSolution = solution.copy();
	
	
		int tempindex = (int)solution.getAttribute(CROWDING_NUMBER);
		newSolution.setAttribute(CROWDING_NUMBER, tempindex);
		double tempdis = (double)solution.getAttribute(DISTANCE_ATTRIBUTE);
		newSolution.setAttribute(DISTANCE_ATTRIBUTE, tempdis);
		double tempCdis = (double)solution.getAttribute(CROWDING_ATTRIBUTE);
		newSolution.setAttribute(CROWDING_ATTRIBUTE, tempCdis);
		int rank= (int) solution.getAttribute(RANK_ATTRIBUTE);
		newSolution.setAttribute(RANK_ATTRIBUTE, rank);
		
		for (int i = 0; i < solution.getNumberOfObjectives(); i++)
		{
			newSolution.setObjective(i, TRFunctionReverse(solution.getObjective(i), i, minBound, maxBound));
		}
				
	return newSolution;
}


private double FindSmallestInDoubleArray(double[] newdata) {
	 // Validates input
    if (newdata == null) {
        throw new IllegalArgumentException("The Array must not be null");
    } else if (newdata.length == 0) {
        throw new IllegalArgumentException("Array cannot be empty.");
    }

    // Finds and returns min
    double min = newdata[0];
    for (int i = 1; i < newdata.length; i++) {
        if (Double.isNaN(newdata[i])) {
            return Double.NaN;
        }
        if (newdata[i] < min) {
            min = newdata[i];
        }
    }

    return min;
}

private double FindLargestInDoubleArray(double[] newdata) {
	 // Validates input
    if (newdata == null) {
        throw new IllegalArgumentException("The Array must not be null");
    } else if (newdata.length == 0) {
        throw new IllegalArgumentException("Array cannot be empty.");
    }

    // Finds and returns max
    double max = newdata[0];
    for (int i = 1; i < newdata.length; i++) {
        if (Double.isNaN(newdata[i])) {
            return Double.NaN;
        }
        if (newdata[i] > max) {
       	 max = newdata[i];
        }
    }

    return max;
}

private boolean HasNoneZeroValue(double[] crowdingdata) {
	// TODO at least one of the elements in crowdingdata is not 0
	boolean redult=false;
	int num=crowdingdata.length;
	for(int i=0;i<num;i++){
		if(crowdingdata[i]!=0.0)
		{
			redult=true;
			break;
		}
	}
	return redult;
}

	@Override
	public boolean areLargerValuesPreferred() {
		return true;
	}

	@Override
	public void evaluateDF(Population population) {
		if (population.size() <= 2) {
			for (Solution solution : population) {
				solution.setAttribute(FITNESS_ATTRIBUTE, 0.0);
			}
		} else {
			int numberOfObjectives = problem.getNumberOfObjectives();
			List<Solution> solutions = normalizeDF(population);
			List<Solution> solutionsCopy = new ArrayList<Solution>(solutions);
			
			double totalVolume = Hypervolume.calculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives);
			
			for (int i = 0; i < population.size(); i++) {
				solutionsCopy = new ArrayList<Solution>(solutions);
				solutionsCopy.remove(i);
				
				double volume = Hypervolume.calculateHypervolume(solutionsCopy, solutionsCopy.size(), numberOfObjectives);
				double hvContribution=totalVolume - volume;
				population.get(i).setAttribute(FITNESS_ATTRIBUTE, hvContribution);
			}
		}
		
	}

	private Solution DFTransformation(Solution s, double[] lsl, double[] usl, double[] d) {
		Solution newSolution = s.copy();
		
		
		int tempindex = (int)s.getAttribute(CROWDING_NUMBER);
		newSolution.setAttribute(CROWDING_NUMBER, tempindex);
//		double tempdis = (double)s.getAttribute(DISTANCE_ATTRIBUTE);
//		newSolution.setAttribute(DISTANCE_ATTRIBUTE, tempdis);
//		double tempCdis = (double)s.getAttribute(CROWDING_ATTRIBUTE);
//		newSolution.setAttribute(CROWDING_ATTRIBUTE, tempCdis);
		int rank= (int) s.getAttribute(RANK_ATTRIBUTE);
		newSolution.setAttribute(RANK_ATTRIBUTE, rank);
		
		for (int i = 0; i < s.getNumberOfObjectives(); i++)
		{
//			newSolution.setObjective(i, -TwoSideDFFunction(s.getObjective(i),  lsl[i], usl[i]));
			newSolution.setObjective(i, -OneSideDFFunction(s.getObjective(i), lsl[i], usl[i],d[i]));
		}
				
	return newSolution;
	}
	
	//Harrington's one-sided DF	
private double OneSideDFFunction(double objective,  double lower, double upper, double d) {
		double result=0;
		double y1=lower;
		double d1=0.999;
		double y2=upper;
		double d2=d;
		double b1=(-Math.log(-Math.log(d2))+ Math.log(-Math.log(d1)))/(y2-y1);
		double b0=-Math.log(-Math.log(d1))-b1*y1;
		result=Math.exp(-Math.exp(-(b0+b1*objective)));
		
		return result;
	}

	//Harrington's two-sided DF
	private double TwoSideDFFunction(double objective, double lsl, double usl) {
		double result=0;
		double y=(2*objective-(lsl+usl))/(usl-lsl);
		result=Math.exp(-Math.pow(Math.abs(y),n));
		
		return result;
	}

}
